<?php
namespace WowGmbh\AmazonPayment\Service;

/***************************************************************
 *
 *  Copyright notice
 *
 *  (c) 2017 WOW-GmbH <info@wow-gmbh.de>, WOW-GmbH.de
 *
 *  All rights reserved
 *
 *  This script is part of the TYPO3 project. The TYPO3 project is
 *  free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  The GNU General Public License can be found at
 *  http://www.gnu.org/copyleft/gpl.html.
 *
 *  This script is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  This copyright notice MUST APPEAR in all copies of the script!
 ***************************************************************/

use \TYPO3\CMS\Core\Utility\GeneralUtility;
use \WowGmbh\AmazonPayment\Domain\Model\Payment;
use \WowGmbh\AmazonPayment\Utility\Tools;

/**
 * Payment service
 *
 * @package amazon_payment
 *
 * @author WOW-GmbH <info@wow-gmbh.de>
 */
class PaymentService
{

    /**
     * @var \PayWithAmazon\Client
     */
    protected $client;

    /**
     * @var \WowGmbh\AmazonPayment\Domain\Model\Payment
     */
    protected $payment;

    /**
     * @var \WowGmbh\AmazonPayment\Domain\Repository\LogRepository
     * @inject
     */
    protected $logRepository;

    /**
     * @var \WowGmbh\AmazonPayment\Domain\Repository\PaymentRepository
     * @inject
     */
    protected $paymentRepository;

    /**
     * persistence manager
     *
     * @var \TYPO3\CMS\Extbase\Persistence\PersistenceManagerInterface
     * @inject
     */
    protected $persistenceManager;

    /**
     * Initializes this service.
     * When there's no Payment record related to the order reference id, it will create one.
     *
     * @param array $config
     * @param string $orderReferenceId
     * @param array $proxy
     * @throws \Exception
     */
    public function initialize($config = array(), $orderReferenceId, $proxy=array())
    {
        $newRecordPersisted = false;
        // Inject client
        if (!class_exists('\PayWithAmazon\Client')) {
            $classPath = 'EXT:amazon_payment/Resources/Private/Php/login-and-pay-with-amazon-sdk-php/';
            $classPath = GeneralUtility::getFileAbsFileName($classPath . 'PayWithAmazon/Client.php');
            require_once $classPath;
        }
        $this->client = new \PayWithAmazon\Client($config);
        // Use proxy
        if(!empty($proxy)) {
            $this->client->setProxy($proxy);
        }
        //
        // Get the related payment object
        $this->payment = $this->paymentRepository->findOneByAmazonOrderReferenceId($orderReferenceId);
        if ($this->payment instanceof Payment) {
            // Payment is already persisted? Fine!
        } // When there is no payment object yet, create one
        else {
            /** @var \WowGmbh\AmazonPayment\Domain\Model\Log $log */
            $this->payment = GeneralUtility::makeInstance('WowGmbh\AmazonPayment\Domain\Model\Payment');
            // General data
            $this->payment->setAmazonOrderReferenceId($orderReferenceId);
            $this->paymentRepository->add($this->payment);
            $this->persistenceManager->persistAll();
            $newRecordPersisted = true;
        }
        // Be sure, the payment object is really available!
        if (!($this->payment instanceof Payment)) {
            throw new \Exception('Payment object could not be initialized');
        }
        return $newRecordPersisted;
    }

    /**
     * @param $action string
     * @param $request array
     * @param $response array
     */
    protected function log($action, $request, $response)
    {
        /** @var \WowGmbh\AmazonPayment\Domain\Model\Log $log */
        $log = GeneralUtility::makeInstance('WowGmbh\AmazonPayment\Domain\Model\Log');
        // General data
        $orderReferenceId = $this->payment->getAmazonOrderReferenceId();
        if (isset($request['amazon_order_reference_id'])) {
            $orderReferenceId = $request['amazon_order_reference_id'];
        }
        $log->setAmazonOrderReferenceId($orderReferenceId);
        $log->setAction($action);
        $log->setPayment($this->payment);
        // Request
        $log->setRequest(serialize($request));
        $log->setRequestText(Tools::arrayToString($request));
        // Response
        $log->setResponseStatus($response['ResponseStatus']);
        $log->setResponse(serialize($response));
        $log->setResponseText(Tools::arrayToString($response));
        // Write into log
        $this->logRepository->add($log);
    }

    /**
     * Update payment object
     * @param $response
     * @return mixed
     */
    protected function updatePayment($response)
    {
        // Get order reference details
        if (isset($response['GetOrderReferenceDetailsResult'])) {
            $getOrderReferenceDetailsResult = $response['GetOrderReferenceDetailsResult'];
            if (isset($getOrderReferenceDetailsResult['OrderReferenceDetails'])) {
                $orderReferenceDetails = $getOrderReferenceDetailsResult['OrderReferenceDetails'];
                if (isset($orderReferenceDetails['OrderReferenceStatus'])) {
                    $orderReferenceStatus = $orderReferenceDetails['OrderReferenceStatus'];
                    // Update order state, if available
                    if (isset($orderReferenceStatus['State'])) {
                        $this->payment->setAmazonOrderReferenceStatus($orderReferenceStatus['State']);
                    }
                }
                if (isset($orderReferenceDetails['OrderTotal'])) {
                    $orderTotal = $orderReferenceDetails['OrderTotal'];
                    // Update currency and amount, if available
                    if (isset($orderTotal['CurrencyCode'])) {
                        $this->payment->setCurrencyCode($orderTotal['CurrencyCode']);
                    }
                    if (isset($orderTotal['Amount'])) {
                        $this->payment->setAmount($orderTotal['Amount']);
                    }
                }
            }
        }
        // Get user info
        if (isset($response['user_id'])) {
            $this->payment->setAmazonUserId($response['user_id']);
            // Set the response status to 200 (OK)
            $response['ResponseStatus'] = 200;
        }
        if (isset($response['name'])) {
            $this->payment->setAmazonUsername($response['name']);
        }
        if (isset($response['email'])) {
            $this->payment->setAmazonEmail($response['email']);
        }
        // If payment object is dirty, update it in repository
        if ($this->payment->_isDirty()) {
            $this->paymentRepository->update($this->payment);
        }
        return $response;
    }

    /**
     * @param $prefix string Prefix for order is
     * @return string
     */
    public function getPaymentOrderId($prefix)
    {
        return $prefix . (string)$this->payment->getUid() . '-' . time();
    }

    /**
     * @return string
     */
    public function getAmazonOrderReferenceStatus()
    {
        return $this->payment->getAmazonOrderReferenceStatus();
    }

    /**
     * @return string
     */
    public function getAmount()
    {
        return $this->payment->getAmount();
    }

    /**
     * @return boolean
     */
    public function isAddressRequired()
    {
        return $this->payment->isAddressRequired();
    }

    /**
     * @return boolean
     */
    public function isChargeOnShipment()
    {
        return $this->payment->isChargeOnShipment();
    }

    /**
     * @return string
     */
    public function getCurrencyCode()
    {
        return $this->payment->getCurrencyCode();
    }

    /**
     * GetUserInfo convenience function
     * Returns user's profile information from Amazon using the access token returned by the Button widget.
     *
     * @param string $accessToken
     * @return array
     */
    public function getUserInfo($accessToken)
    {
        try {
            $response = $this->client->getUserInfo($accessToken);
            $response = $this->updatePayment($response);
            $this->log('getUserInfo', array($accessToken), $response);
        } catch (\Exception $e) {
            $response = array();
            $response['ResponseStatus'] = 500;
            $response['Error']['Code'] = $e->getCode();
            $response['Error']['Message'] = $e->getMessage();
        }
        return $response;
    }

    /**
     * GetOrderReferenceDetails API call
     * Returns details about the Order Reference object and its current state.
     *
     * @param string $requestParameters ['merchant_id'] - [String]
     * @param requestParameters ['amazon_order_reference_id'] - [String]
     * @optional requestParameters['address_consent_token'] - [String]
     * @optional requestParameters['mws_auth_token'] - [String]
     * @return array
     */
    public function getOrderReferenceDetails($requestParameters = array())
    {
        $response = $this->client->getOrderReferenceDetails($requestParameters);
        $response = $response->toArray();
        $response = $this->updatePayment($response);
        $this->log('getOrderReferenceDetails', $requestParameters, $response);
        return $response;
    }

    /**
     * SetOrderReferenceDetails API call
     * Sets order reference details such as the order total and a description for the order.
     *
     * @param requestParameters ['merchant_id'] - [String]
     * @param requestParameters ['amazon_order_reference_id'] - [String]
     * @param requestParameters ['amount'] - [String]
     * @param requestParameters ['currency_code'] - [String]
     * @optional requestParameters['platform_id'] - [String]
     * @optional requestParameters['seller_note'] - [String]
     * @optional requestParameters['seller_order_id'] - [String]
     * @optional requestParameters['store_name'] - [String]
     * @optional requestParameters['custom_information'] - [String]
     * @optional requestParameters['mws_auth_token'] - [String]
     * @return array
     */
    public function setOrderReferenceDetails($requestParameters = array(), \WowGmbh\AmazonPayment\Domain\Model\PaymentPreset $paymentPreset)
    {
        $response = $this->client->setOrderReferenceDetails($requestParameters);
        $response = $response->toArray();
        // Set address-required and charge-on-shipment
        $this->payment->setAddressRequired($paymentPreset->isAddressRequired());
        $this->payment->setChargeOnShipment($paymentPreset->isChargeOnShipment());
        $response = $this->updatePayment($response);
        $this->log('setOrderReferenceDetails', $requestParameters, $response);
        return $response;
    }

    /**
     * ConfirmOrderReferenceDetails API call
     * Confirms that the order reference is free of constraints and
     * all required information has been set on the order reference.
     *
     * @param requestParameters ['merchant_id'] - [String]
     * @param requestParameters ['amazon_order_reference_id'] - [String]
     * @optional requestParameters['mws_auth_token'] - [String]
     */
    public function confirmOrderReference($requestParameters = array())
    {
        $response = $this->client->confirmOrderReference($requestParameters);
        $response = $response->toArray();
        if (!isset($response['Error'])) {
            // Set payment record as open
            $response['GetOrderReferenceDetailsResult']['OrderReferenceDetails']['OrderReferenceStatus']['State'] = 'Open';
        }
        $response = $this->updatePayment($response);
        $this->log('confirmOrderReference', $requestParameters, $response);
        return $response;
    }

    /* CancelOrderReferenceDetails API call - Cancels a previously confirmed order reference.
     * @see http://docs.developer.amazonservices.com/en_US/off_amazon_payments/OffAmazonPayments_CancelOrderReference.html
     *
     * @param requestParameters['merchant_id'] - [String]
     * @param requestParameters['amazon_order_reference_id'] - [String]
     * @optional requestParameters['cancelation_reason'] [String]
     * @optional requestParameters['mws_auth_token'] - [String]
     */

    public function cancelOrderReference($requestParameters = array())
    {
        $response = $this->client->cancelOrderReference($requestParameters);
        $response = $response->toArray();
        $response = $this->updatePayment($response);
        $this->log('cancelOrderReference', $requestParameters, $response);
        return $response;
    }

    /**
     * Authorize API call
     * Reserves a specified amount against the payment method(s) stored in the order reference.
     *
     * @param requestParameters ['merchant_id'] - [String]
     * @param requestParameters ['amazon_order_reference_id'] - [String]
     * @param requestParameters ['authorization_amount'] [String]
     * @param requestParameters ['currency_code'] - [String]
     * @param requestParameters ['authorization_reference_id'] [String]
     * @optional requestParameters['capture_now'] [String]
     * @optional requestParameters['provider_credit_details'] - [array (array())]
     * @optional requestParameters['seller_authorization_note'] [String]
     * @optional requestParameters['transaction_timeout'] [String] - Defaults to 1440 minutes
     * @optional requestParameters['soft_descriptor'] - [String]
     * @optional requestParameters['mws_auth_token'] - [String]
     * @return array
     */
    public function authorize($requestParameters = array())
    {
        $response = $this->client->authorize($requestParameters);
        $response = $response->toArray();
        $response = $this->updatePayment($response);
        $this->log('authorize', $requestParameters, $response);
        return $response;
    }

    /**
     * GetAuthorizationDetails API call
     * Returns the status of a particular authorization and the total amount captured on the authorization.
     *
     * @param requestParameters ['merchant_id'] - [String]
     * @param requestParameters ['amazon_authorization_id'] [String]
     * @optional requestParameters['mws_auth_token'] - [String]
     * @return array
     */
    public function getAuthorizationDetails($requestParameters = array())
    {
        $response = $this->client->getAuthorizationDetails($requestParameters);
        $response = $response->toArray();
        $response = $this->updatePayment($response);
        $this->log('getAuthorizationDetails', $requestParameters, $response);
        return $response;
    }


    /**
     * charge convenience method
     * Performs the API calls
     * 1. SetOrderReferenceDetails / SetBillingAgreementDetails
     * 2. ConfirmOrderReference / ConfirmBillingAgreement
     * 3. Authorize (with Capture) / AuthorizeOnBillingAgreeemnt (with Capture)
     *
     * @param requestParameters ['merchant_id'] - [String]
     *
     * @param requestParameters ['amazon_reference_id'] - [String] : Order Reference ID /Billing Agreement ID
     * If requestParameters['amazon_reference_id'] is empty then the following is required,
     * @param requestParameters ['amazon_order_reference_id'] - [String] : Order Reference ID
     * or,
     * @param requestParameters ['amazon_billing_agreement_id'] - [String] : Billing Agreement ID
     *
     * @param $requestParameters ['charge_amount'] - [String] : Amount value to be captured
     * @param requestParameters ['currency_code'] - [String] : Currency Code for the Amount
     * @param requestParameters ['authorization_reference_id'] - [String]- Any unique string that needs to be passed
     * @optional requestParameters['charge_note'] - [String] : Seller Note sent to the buyer
     * @optional requestParameters['transaction_timeout'] - [String] : Defaults to 1440 minutes
     * @optional requestParameters['charge_order_id'] - [String] : Custom Order ID provided
     * @optional requestParameters['mws_auth_token'] - [String]
     */
    public function charge($requestParameters = array())
    {
        $response = $this->client->charge($requestParameters);
        $response = $response->toArray();
        $response = $this->updatePayment($response);
        $this->log('charge', $requestParameters, $response);
        return $response;
    }

    /**
     * CloseOrderReferenceDetails API call
     * Confirms that an order reference has been fulfilled (fully or partially)
     * and that you do not expect to create any new authorizations on this order reference.
     * @see http://docs.developer.amazonservices.com/en_US/off_amazon_payments/OffAmazonPayments_CloseOrderReference.html
     *
     * @param requestParameters ['merchant_id'] - [String]
     * @param requestParameters ['amazon_order_reference_id'] - [String]
     * @optional requestParameters['closure_reason'] [String]
     * @optional requestParameters['mws_auth_token'] - [String]
     */

    public function closeOrderReference($requestParameters = array())
    {
        $response = $this->client->closeOrderReference($requestParameters);
        $response = $response->toArray();
        // Set payment record as closed
        $response['GetOrderReferenceDetailsResult']['OrderReferenceDetails']['OrderReferenceStatus']['State'] = 'Closed';
        $response = $this->updatePayment($response);
        $this->log('closeOrderReference', $requestParameters, $response);
        return $response;
    }

    /**
     * Extracts order reference details for frontend rendering
     * @param $orderReferenceDetails
     * @return array
     */
    public function extractOrderReferenceDetails($orderReferenceDetails, $debug=false)
    {
        $data = array();
        /**
         * Order state:
         * - Draft
         * - Closed
         * - Open
         */
        $data['orderState'] = $orderReferenceDetails['OrderReferenceStatus']['State'];
        $data['sellerOrderAttributes'] = array();
        $data['sellerOrderAttributes']['storeName'] = $orderReferenceDetails['SellerOrderAttributes']['StoreName'];
        $data['sellerOrderAttributes']['customInformation'] = $orderReferenceDetails['SellerOrderAttributes']['CustomInformation'];
        $data['sellerOrderAttributes']['sellerOrderId'] = $orderReferenceDetails['SellerOrderAttributes']['SellerOrderId'];
        $data['orderTotal'] = array();
        $data['orderTotal']['currencyCode'] = $orderReferenceDetails['OrderTotal']['CurrencyCode'];
        $data['orderTotal']['amount'] = $orderReferenceDetails['OrderTotal']['Amount'];
        $data['sellerNote'] = $orderReferenceDetails['SellerNote'];
        $data['buyer'] = array();
        $data['buyer']['name'] = $orderReferenceDetails['Buyer']['Name'];
        $data['buyer']['email'] = $orderReferenceDetails['Buyer']['Email'];
        // Shipping address
        if (isset($orderReferenceDetails['Destination'])) {
            $data['shippingAddress'] = array();
            $data['shippingAddress']['phone'] = $orderReferenceDetails['Destination']['PhysicalDestination']['Phone'];
            $data['shippingAddress']['city'] = $orderReferenceDetails['Destination']['PhysicalDestination']['City'];
            $data['shippingAddress']['countryCode'] = $orderReferenceDetails['Destination']['PhysicalDestination']['CountryCode'];
            $data['shippingAddress']['postalCode'] = $orderReferenceDetails['Destination']['PhysicalDestination']['PostalCode'];
            $data['shippingAddress']['name'] = $orderReferenceDetails['Destination']['PhysicalDestination']['Name'];
            $data['shippingAddress']['address'] = $orderReferenceDetails['Destination']['PhysicalDestination']['AddressLine2'];
        }
        // Billing address
        if (isset($orderReferenceDetails['BillingAddress'])) {
            $data['billingAddress'] = array();
            $data['billingAddress']['city'] = $orderReferenceDetails['BillingAddress']['PhysicalAddress']['City'];
            $data['billingAddress']['countryCode'] = $orderReferenceDetails['BillingAddress']['PhysicalAddress']['CountryCode'];
            $data['billingAddress']['postalCode'] = $orderReferenceDetails['BillingAddress']['PhysicalAddress']['PostalCode'];
            $data['billingAddress']['name'] = $orderReferenceDetails['BillingAddress']['PhysicalAddress']['Name'];
            $data['billingAddress']['address'] = $orderReferenceDetails['BillingAddress']['PhysicalAddress']['AddressLine2'];
        }
        // Constraints
        if (isset($orderReferenceDetails['Constraints'])) {
            $data['constraints'] = array();
            // Only one constraint
            if (isset($orderReferenceDetails['Constraints']['Constraint']['ConstraintID'])) {
                $data['constraints'][$orderReferenceDetails['Constraints']['Constraint']['ConstraintID']] =
                    Tools::processConstraintMessage($orderReferenceDetails['Constraints']['Constraint']['ConstraintID'], $orderReferenceDetails['Constraints']['Constraint']['Description'], $debug);
            } // Multiple constraints
            else {
                foreach ($orderReferenceDetails['Constraints']['Constraint'] as $constraint) {
                    $data['constraints'][$constraint['ConstraintID']] = Tools::processConstraintMessage($constraint['ConstraintID'], $constraint['Description'], $debug);
                }
            }
        }
        return $data;
    }

    /**
     * Extracts the authorizaton details
     * @param $authorizationDetails
     * @return array
     */
    public function extractAuthorizationDetails($authorizationDetails)
    {
        $data = array();
        if (isset($authorizationDetails['AuthorizationAmount'])) {
            $data['authorizationAmount'] = array();
            $data['authorizationAmount']['currencyCode'] = $authorizationDetails['AuthorizationAmount']['CurrencyCode'];
            $data['authorizationAmount']['amount'] = $authorizationDetails['AuthorizationAmount']['Amount'];
        }
        if (isset($authorizationDetails['AuthorizationStatus'])) {
            $data['authorizationStatus'] = array();
            $data['authorizationStatus']['reasonCode'] = $authorizationDetails['AuthorizationStatus']['ReasonCode'];
            $data['authorizationStatus']['state'] = $authorizationDetails['AuthorizationStatus']['State'];
        }
        return $data;
    }
}

?>